# -*- coding: utf-8 -*-
"""
util.py

Created by Stephan Hügel on 2015-06-22

This file is part of convertbng.

"""

import os
from array import array
from ctypes import (
    Structure,
    c_double,
    c_size_t,
    c_void_p,
    cast,
    cdll,
    string_at,
)
from sys import platform, version_info

import numpy as np

__author__ = "Stephan Hügel"

file_path = os.path.dirname(__file__)

if platform == "darwin":
    prefix = "lib"
    ext = "dylib"
elif "linux" in platform:
    prefix = "lib"
    ext = "so"
    fpath = os.path.join(file_path, ".libs")

elif "win32" in platform:
    prefix = ""
    ext = "dll"

prefix = {"win32": ""}.get(platform, "lib")
extension = {"darwin": ".dylib", "win32": ".dll"}.get(platform, ".so")

# Python 3 check
if version_info > (3, 0):
    from subprocess import getoutput as spop

    py3 = True
else:
    from subprocess import check_output as spop

    py3 = False

try:
    lib = cdll.LoadLibrary(os.path.join(file_path, prefix + "lonlat_bng" + extension))
except OSError:
    # the Rust lib's been grafted by manylinux1
    if not py3:
        fname = spop(["ls", fpath]).split()[0]
    else:
        fname = spop(["ls %s" % fpath]).split()[0]
    lib = cdll.LoadLibrary(os.path.join(file_path, ".libs", fname))


class _FFIArray(Structure):
    """Convert sequence of floats to a C-compatible void array"""

    _fields_ = [("data", c_void_p), ("len", c_size_t)]

    @classmethod
    def from_param(cls, seq):
        """Allow implicit conversions from a sequence of 64-bit floats."""
        return seq if isinstance(seq, cls) else cls(seq)

    def __init__(self, seq, data_type=c_double):
        """
        Convert sequence of values into array, then ctypes Structure

        Rather than checking types (bad), we just try to blam seq
        into a ctypes object using from_buffer. If that doesn't work,
        we try successively more conservative approaches:
        numpy array -> array.array -> read-only buffer -> CPython iterable
        """
        if isinstance(seq, float):
            seq = array("d", [seq])
        try:
            len(seq)
        except TypeError:
            # we've got an iterator or a generator, so consume it
            seq = array("d", seq)
        array_type = data_type * len(seq)
        try:
            raw_seq = array_type.from_buffer(seq.astype(np.float64))
        except (TypeError, AttributeError):
            try:
                raw_seq = array_type.from_buffer_copy(seq.astype(np.float64))
            except (TypeError, AttributeError):
                # it's a list or a tuple
                raw_seq = array_type.from_buffer(array("d", seq))
        self.data = cast(raw_seq, c_void_p)
        self.len = len(seq)


class _Result_Tuple(Structure):
    """Container for returned FFI data"""

    _fields_ = [("e", _FFIArray), ("n", _FFIArray)]


def _void_array_to_list(restuple, _func, _args):
    """Convert the FFI result to Python data structures"""
    shape = (restuple.e.len, 1)
    array_size = np.prod(shape)
    mem_size = 8 * array_size

    array_str_e = string_at(restuple.e.data, mem_size)
    array_str_n = string_at(restuple.n.data, mem_size)

    ls_e = np.frombuffer(array_str_e, float, array_size).tolist()
    ls_n = np.frombuffer(array_str_n, float, array_size).tolist()

    return ls_e, ls_n


# Multi-threaded FFI functions
convert_bng = lib.convert_to_bng_threaded
convert_bng.argtypes = (_FFIArray, _FFIArray)
convert_bng.restype = _Result_Tuple
convert_bng.errcheck = _void_array_to_list
convert_bng.__doc__ = """
    Multi-threaded lon, lat --> BNG conversion
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    Uses the Helmert transform
    """

convert_lonlat = lib.convert_to_lonlat_threaded
convert_lonlat.argtypes = (_FFIArray, _FFIArray)
convert_lonlat.restype = _Result_Tuple
convert_lonlat.errcheck = _void_array_to_list
convert_lonlat.__doc__ = """
    Multi-threaded BNG --> lon, lat conversion
    Returns a list of two lists containing Longitude and Latitude floats,
    respectively
    Uses the Helmert transform
    """

convert_to_osgb36 = lib.convert_to_osgb36_threaded
convert_to_osgb36.argtypes = (_FFIArray, _FFIArray)
convert_to_osgb36.restype = _Result_Tuple
convert_to_osgb36.errcheck = _void_array_to_list
convert_to_osgb36.__doc__ = """
    Multi-threaded lon, lat --> OSGB36 conversion, using OSTN02 data
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    """

convert_osgb36_to_lonlat = lib.convert_osgb36_to_ll_threaded
convert_osgb36_to_lonlat.argtypes = (_FFIArray, _FFIArray)
convert_osgb36_to_lonlat.restype = _Result_Tuple
convert_osgb36_to_lonlat.errcheck = _void_array_to_list
convert_osgb36_to_lonlat.__doc__ = """
    Multi-threaded OSGB36 --> Lon, Lat conversion, using OSTN02 data
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    """

convert_etrs89_to_lonlat = lib.convert_etrs89_to_ll_threaded
convert_etrs89_to_lonlat.argtypes = (_FFIArray, _FFIArray)
convert_etrs89_to_lonlat.restype = _Result_Tuple
convert_etrs89_to_lonlat.errcheck = _void_array_to_list
convert_etrs89_to_lonlat.__doc__ = """
    Multi-threaded ETRS89 Eastings and Northings --> OSGB36 conversion, using OSTN02 data
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    """

convert_etrs89_to_osgb36 = lib.convert_etrs89_to_osgb36_threaded
convert_etrs89_to_osgb36.argtypes = (_FFIArray, _FFIArray)
convert_etrs89_to_osgb36.restype = _Result_Tuple
convert_etrs89_to_osgb36.errcheck = _void_array_to_list
convert_etrs89_to_osgb36.__doc__ = """
    Multi-threaded OSGB36 Eastings and Northings --> ETRS89 Eastings and Northings conversion,
    using OSTN02 data
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    """

convert_osgb36_to_etrs89 = lib.convert_osgb36_to_etrs89_threaded
convert_osgb36_to_etrs89.argtypes = (_FFIArray, _FFIArray)
convert_osgb36_to_etrs89.restype = _Result_Tuple
convert_osgb36_to_etrs89.errcheck = _void_array_to_list
convert_osgb36_to_etrs89.__doc__ = """
    Multi-threaded ETRS89 Eastings and Northings --> Lon, Lat conversion,
    Returns a list of two lists containing Longitude and Latitude floats,
    respectively
    """

convert_epsg3857_to_wgs84 = lib.convert_epsg3857_to_wgs84_threaded
convert_epsg3857_to_wgs84.argtypes = (_FFIArray, _FFIArray)
convert_epsg3857_to_wgs84.restype = _Result_Tuple
convert_epsg3857_to_wgs84.errcheck = _void_array_to_list
convert_epsg3857_to_wgs84.__doc__ = """
    Convert Google Web Mercator (EPSG3857) coordinates to WGS84
    Latitude and Longitude
    Returns a list of two lists containing latitudes and longitudes,
    respectively
    """

convert_to_etrs89 = lib.convert_to_etrs89_threaded
convert_to_etrs89.argtypes = (_FFIArray, _FFIArray)
convert_to_etrs89.restype = _Result_Tuple
convert_to_etrs89.errcheck = _void_array_to_list
convert_to_etrs89.__doc__ = """
    Transform longitudes and latitudes to ETRS89 Eastings and Northings (without OSTN15 corrections)
    Returns a list of two lists containing Easting and Northing floats,
    respectively
    """

# Free FFI-allocated memory
drop_array = lib.drop_float_array
drop_array.argtypes = (_FFIArray, _FFIArray)
drop_array.restype = None



import array
import math
import unittest
from ctypes import ArgumentError
from random import randint, randrange

import numpy as np
from convertbng.cutil import convert_bng as cconvert_bng

from convertbng.util import (
    convert_bng,
    convert_epsg3857_to_wgs84,
    convert_etrs89_to_lonlat,
    convert_etrs89_to_osgb36,
    convert_lonlat,
    convert_osgb36_to_etrs89,
    convert_osgb36_to_lonlat,
    convert_to_etrs89,
    convert_to_osgb36,
)


class ConvertbngTests(unittest.TestCase):
    """Tests for convert_bng"""

    def setUp(self):
        """Set stuff up"""
        pass

    def testConvertLonLat(self):
        """Test multithreaded lon, lat --> BNG function"""
        expected = ([516274.46, 398915.554], [173141.101, 521544.09])

        result = convert_bng(
            [-0.32824866, -2.0183041005533306], [51.44533267, 54.589097162646141]
        )
        self.assertEqual(expected, result)

    def testCythonConvertLonLat(self):
        """Test Cythonised multithreaded lon, lat --> BNG function"""
        expected = ([516274.46000000002, 398915.554], [173141.101, 521544.09000000003])

        result = cconvert_bng(
            np.array([-0.32824866, -2.0183041005533306]),
            np.array([51.44533267, 54.589097162646141]),
        )
        # dump array result into list
        a, b = list(result[0]), list(result[1])
        self.assertEqual(expected, (a, b))

    def testCythonConvertList(self):
        """Test Cythonised multithreaded lon, lat --> BNG function with lists"""
        expected = (
            np.array([516274.460, 398915.542]),
            np.array([173141.098, 521544.088]),
        )

        result = cconvert_bng(
            [-0.32824866, -2.0183041005533306], [51.44533267, 54.589097162646141]
        )
        self.assertEqual(expected[0][0], result[0][0])

    def test_loop(self):
        """See https://github.com/urschrei/convertbng/issues/101"""

        def test_conversion(easting, northing):
            res = convert_lonlat(easting, northing)
            return [res[0][0], res[1][0]]

        for i in range(500000):
            test_conversion([randint(10000, 50000)], [randint(10000, 50000)])

    def testConvertBNG(self):
        """Test multithreaded BNG --> lon, lat function"""
        expected = ([-0.32822654, -2.01831267], [51.44533145, 54.58910534])
        result = convert_lonlat([516276, 398915], [173141, 521545])
        self.assertEqual(expected, result)

    def testConvertLonLatSingle(self):
        """Test lon, lat --> BNG conversion of single values"""
        expected = ([651409.804], [313177.45])
        result = convert_bng(1.716073973, 52.658007833)
        self.assertEqual(expected, result)

    def testConvertTuple(self):
        """Test lon, lat --> BNG conversion of tuples"""
        expected = ([651409.804], [313177.45])
        result = convert_bng((1.716073973,), (52.658007833,))
        self.assertEqual(expected, result)

    def testConvertString(self):
        """Test that an error is thrown for incorrect types"""
        with self.assertRaises(ArgumentError) as result:
            convert_bng(["Foo"], ["Bar"])

    def testConvertIterable(self):
        """Test lon, lat --> BNG conversion of tuples"""
        expected = ([651409.804], [313177.45])
        result = convert_bng(iter([1.7160739736]), iter([52.658007833]))
        self.assertEqual(expected, result)

    # def testConvertArray(self):
    #     """ Test lon, lat --> BNG conversion of array.array """
    #     expected = ([651409.804], [313177.45])
    #     result = convert_bng(
    #         array.array("d", [1.716073973]), array.array("d", [52.658007833])
    #     )
    #     self.assertEqual(expected, result)

    def testGenerator(self):
        """Test that the lon, lat -> BNG function can consume generators"""
        expected = ([516274.46, 398915.554], [173141.101, 521544.09])
        inp = [[-0.32824866, -2.0183041005533306], [51.44533267, 54.589097162646141]]
        lon_generator = (n for n in inp[0])
        lat_generator = (n for n in inp[1])
        result = convert_bng(lon_generator, lat_generator)
        self.assertEqual(expected, result)

    def testNumpyConversion(self):
        """Test lon, lat --> BNG conversion of numpy arrays"""
        # UK bounding box
        N = 55.811741
        E = 1.768960
        S = 49.871159
        W = -6.379880

        num_coords = 1000
        lon_arr = np.random.uniform(W, E, [num_coords])
        lat_arr = np.random.uniform(S, N, [num_coords])
        convert_bng(lon_arr, lat_arr)

    def testLargeArrayConversion(self):
        """Test that we don't get segmentation fault: 11 on large (1MM points) arrays using Ctypes"""
        # UK bounding box
        N = 55.811741
        E = 1.768960
        S = 49.871159
        W = -6.379880

        num_coords = 1000000
        lon_ls = list(np.random.uniform(W, E, [num_coords]))
        lat_ls = list(np.random.uniform(S, N, [num_coords]))
        convert_bng(lon_ls, lat_ls)

    def testLargeArrayConversionCython(self):
        """Test that we don't get segmentation fault: 11 on large (1MM points) arrays using Cython"""
        # London bounding box
        N = 51.691874116909894
        E = 0.3340155643740321
        S = 51.28676016315085
        W = -0.5103750689005356

        num_coords = 1000000
        lon_ls = np.random.uniform(W, E, [num_coords])
        lat_ls = np.random.uniform(S, N, [num_coords])
        cconvert_bng(lon_ls, lat_ls)

    def testCythonConsistency(self):
        """Ensure Ctypes and Cython give same results"""
        # London bounding box
        N = 51.691874116909894
        E = 0.3340155643740321
        S = 51.28676016315085
        W = -0.5103750689005356
        num_coords = 100

        rand = randrange(0, 100)
        lon_ls = np.random.uniform(W, E, [num_coords])
        lat_ls = np.random.uniform(S, N, [num_coords])
        # result is a list, so convert to an array. Ugh.
        res_ctypes = np.array(convert_bng(lon_ls, lat_ls)[0][rand])
        res_cython = cconvert_bng(lon_ls, lat_ls)[0][rand]

        self.assertEqual(res_ctypes, res_cython)

    # def testBadValues(self):
    #     """ Test that values outside the bounding box return -1, -1 """
    #     bad_coords = [
    #         # Below minimum longitude
    #         [[-6.379881], [49.871159]],
    #         # Below minimum latitude
    #         [[1.768960], [49.871156]],
    #         # Above maximum longitude
    #         [[1.768961], [55.811741]],
    #         # Above maximum latitude
    #         [[1.768961], [55.811742]]
    #     ]
    #     for coord in bad_coords:
    #         result = convert_bng(coord[0], coord[1])
    #         self.assertTrue(math.isnan(result[0][0]))
    #         self.assertTrue(math.isnan(result[1][0]))

    def test_osgb36(self):
        """Tests lon, lat --> OSGB36 conversion"""
        expected = ([651409.804], [313177.45])
        result = convert_to_osgb36(1.716073973, 52.658007833)
        self.assertEqual(expected, result)

    def test_osgb36_to_lonlat(self):
        """Tests OSGB36 --> Lon, Lat conversion"""
        # Not sure why this is being truncated
        # expected = [[1.716073973], [52.658007833]]
        expected = ([1.71607397], [52.65800783])
        result = convert_osgb36_to_lonlat(651409.804, 313177.45)
        self.assertEqual(expected, result)

    def test_etrs89_to_osgb36(self):
        """Tests ETRS89 Eastings, Northings --> OSGB36 conversion"""
        expected = ([651409.804], [313177.45])
        result = convert_etrs89_to_osgb36(651307.003, 313255.686)
        self.assertEqual(expected, result)

    def test_osgb36_to_etrs89(self):
        """Tests OSGB36 --> ETRS89 Eastings, Northings conversion"""
        expected = ([651307.003], [313255.686])
        result = convert_osgb36_to_etrs89(651409.804, 313177.45)
        self.assertEqual(expected, result)

    def test_etrs89_to_lonlat(self):
        """Tests ETRS89 --> Lon, Lat conversion"""
        expected = ([1.71607397], [52.65800783])
        result = convert_etrs89_to_lonlat(651307.003, 313255.686)
        self.assertEqual(expected, result)

    def test_lonlat_to_etrs89(self):
        """Tests Lon, Lat --> ETRS89 conversion"""
        expected = ([651307.003], [313255.686])
        result = convert_to_etrs89(1.71607397, 52.65800783)
        self.assertEqual(expected, result)

    def test_epsg3857_to_wgs84(self):
        """Tests EPSG3857 to WGS84 conversion"""
        expected = ([-5.625000000783013], [52.48278022732355])
        result = convert_epsg3857_to_wgs84(-626172.1357121646, 6887893.4928337997)

    def test_large_array(self):
        """
        Try to trigger segfault as per https://github.com/urschrei/convertbng/issues/9
        Note: this segfault currently only appears when using ctypes

        """
        [cconvert_bng([-1.89983], [52.48142]) for x in range(10000)]



cdef extern from "header.h":
    struct Array:
        void* data
        size_t len

    struct ResultTuple:
        Array e
        Array n

    cdef ResultTuple convert_to_bng_threaded(Array x, Array y)
    cdef ResultTuple convert_to_lonlat_threaded(Array x, Array y);
    cdef ResultTuple convert_to_osgb36_threaded(Array x, Array y);
    cdef ResultTuple convert_to_etrs89_threaded(Array x, Array y);
    cdef ResultTuple convert_etrs89_to_osgb36_threaded(Array x, Array y);
    cdef ResultTuple convert_etrs89_to_ll_threaded(Array x, Array y);
    cdef ResultTuple convert_osgb36_to_ll_threaded(Array x, Array y);
    cdef ResultTuple convert_osgb36_to_etrs89_threaded(Array x, Array y);
    cdef ResultTuple convert_epsg3857_to_wgs84_threaded(Array x, Array y);
    cdef void drop_float_array(Array x, Array y)



#cython: boundscheck=False
# -*- coding: utf-8 -*-
"""
cutil.pyx

Created by Stephan Hügel on 2015-06-22

This file is part of convertbng.
"""
__author__ = u"Stephan Hügel"

import numpy as np
from convertbng_p cimport (
    Array,
    ResultTuple,
    convert_to_bng_threaded,
    convert_to_lonlat_threaded,
    convert_to_osgb36_threaded,
    convert_to_etrs89_threaded,
    convert_etrs89_to_osgb36_threaded,
    convert_etrs89_to_ll_threaded,
    convert_osgb36_to_ll_threaded,
    convert_osgb36_to_etrs89_threaded,
    convert_epsg3857_to_wgs84_threaded,
    drop_float_array
    )

# should catch a ValueError here, in case something non-double is passed
cdef ffi_wrapper(ResultTuple (*func)(Array, Array)):
    """ Accepts a function pointer and two arguments
            longitudes
            latitudes
    These must be memory-contiguous Python arrays
    The function pointer must map to an external FFI Rust function
    """
    def wrapped(inlon, inlat):
        # Assumes that the data is double, not float. This is easily changed
        # The [::1] promises it's contiguous in memory
        cdef double[::1] wlon = np.array(inlon, dtype=np.float64)
        cdef double [::1] wlat = np.array(inlat, dtype=np.float64)
        cdef Array x_ffi, y_ffi
        # get a pointer to the data, and cast it to void*
        x_ffi.data = <void*>&wlon[0]
        # This may be ... * sizeof(double) - it depends on the C api
        x_ffi.len = wlon.shape[0]
        # repeat for lats
        y_ffi.data = <void*>&wlat[0]
        y_ffi.len = wlat.shape[0]
        # call across the FFI boundary
        cdef ResultTuple result = func(x_ffi, y_ffi)
        # get data pointers for the two result arrays
        cdef double* eastings_ptr = <double*>(result.e.data)
        cdef double* northings_ptr = <double*>(result.n.data)
        # now view the output arrays using memoryviews
        # their length must be specified
        cdef double[::1] e = <double[:result.e.len:1]>eastings_ptr
        cdef double[::1] n = <double[:result.n.len:1]>northings_ptr
        # create numpy copies of the two arrays
        e_numpy = np.copy(e)
        n_numpy = np.copy(n)
        return e_numpy, n_numpy
    return wrapped

convert_bng = ffi_wrapper(&convert_to_bng_threaded)
convert_lonlat = ffi_wrapper(&convert_to_lonlat_threaded)
convert_to_osgb36 = ffi_wrapper(&convert_to_osgb36_threaded)
convert_to_etrs89 = ffi_wrapper(&convert_to_etrs89_threaded)
convert_etrs89_to_osgb36 = ffi_wrapper(&convert_etrs89_to_osgb36_threaded)
convert_etrs89_to_ll = ffi_wrapper(&convert_etrs89_to_ll_threaded)
convert_osgb36_to_ll = ffi_wrapper(&convert_osgb36_to_ll_threaded)
convert_osgb36_to_etrs89 = ffi_wrapper(&convert_osgb36_to_etrs89_threaded)
convert_epsg3857_to_wgs84 = ffi_wrapper(&convert_epsg3857_to_wgs84_threaded)
